///////////////////////////////////////////////////////////////////////////////
//
//  Filename: BCommon.h
//  Author: Wang, Yi
//  Last modified: 2004-2-24 14:29
//
///////////////////////////////////////////////////////////////////////////////

#ifndef BCommonH
#define BCommonH

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stream.h>
#include <fstream.h>

extern const double pi;
extern const double _2pi;

// struct of the parameters
struct tagPara
{
	double MinR;
	double MaxR;
	double PlusA; // plus acceleration
	double NegA; // negative acceleration
	double MaxSpeed;
	double SpeedFact; // factor of speed: SpeedF*speed=CommandValue of the speed
	double LenkFact; // factor of Lenkung: LenkF*lenk=CommandValue of the Lenkung
	double MinSpeedHold; // min length of the distance to hold a speed
};

struct tagCircle
{
	double ox;
	double oy;
	double r;
};

struct tagArc
{
	tagCircle c; // circle of the arc
	double a;  // start angle
	double alen; // angle length (+: counter-clockwise, -: clockwise)
};

struct tagVect
{
	double x;
	double y;
	double alpha;
};

struct tagPos
{
	double time;
	double speed;
	tagVect vect;
};

// struct of design point
struct tagBahnDesignPt
{
	tagVect Vect;
	int time;
	double Speed;
	int StkIdSt;  // index of the first Strecke
	int StkCount; // count of Stks between this and the next DesignPt
	int SpdIdSt;  // index of the first item in SpeedList
	int SpdCount; // count of items in SpeedList between this and the next DesignPt
};

struct tagBahnStrecke
{
    int Type; // 0: Line; 1: Circle
    tagVect VectSt;
    tagVect VectEd;
    double length;
	tagArc arc;
	double TimeSt;
};

// struct of control points
struct tagCtlPoint 
{
	double time; // timestamp
	int Type; // 0: Lenk; 1: Speed;
	double Value; // value of Lenkung or Speed
};

struct tagCalcGeoIContext
{
	tagBahnDesignPt *p1;  
	tagBahnDesignPt *p2;
	tagBahnStrecke stk;
	tagVect vc; // CrossVect
	double d1, d2, dmin;
	bool isParallel;
	int CrossMode;
	// when far
	// tangent circles:  cl1, cr1, cl2, cr2
	tagCircle c[4]; 
	int ci1,ci2;
	double dist_c1_c2;
	tagArc arc1, arc2;
	tagVect tv1, tv2;
};

struct tagSpeedListContext
{
	bool IsErr;
	int mode;
	double tSt;
	double vSt, vEd, dtSt, dtEd;
	double dTime, tLen;
	double au, ad, tu, td;
	double sMax, sMin, vMax, vMin;
};

struct tagSpeedListItem
{
	int BDPtId; // index of DesignPt
	int mode;
	double t0; // timestamp
	double a;  // accelerate
	double v0; // v0
	double l;  // length
	double dt; // duration
};

double getDist(double x1, double y1, double x2, double y2);
double _fmod(double x, double a);
void getVectTanCircle(tagVect &v, tagCircle &cl, tagCircle &cr, double r);
bool getCrossVect(tagVect &v1, tagVect &v2, tagVect &vc);

extern tagPara m_Para;

// for debug
extern ofstream m_Log;

void _dump(tagVect &vect, int id);
void _dump(tagCircle &circle, int id);
void _dump(tagArc &arc, int id);
void _dump(tagBahnDesignPt &BDPt, int id);
void _dump(tagBahnStrecke &stk, int id);
void _dump(tagCtlPoint &cpt, int id);
void _dump(tagSpeedListItem &spli, int id);

#endif
