///////////////////////////////////////////////////////////////////////////////
//
//  Filename: BahnDesign.cpp
//  Author: Wang, Yi
//  Last modified: 2004-2-24 14:30
//
///////////////////////////////////////////////////////////////////////////////

#include "BahnDesign.h"

/***********************  Global  *********************************************/

/*----------------------------------------------------------------------------*/

BahnDesign::BahnDesign(void)
{
    Count = 0;
    StkCt = 0;
    
    // init parameters
	m_Para.MinR = 10;
	m_Para.MaxR = 15;
	m_Para.PlusA = 4;
	m_Para.NegA = 4;
	m_Para.MaxSpeed = 100;
	m_Para.SpeedFact = 100;
	m_Para.LenkFact = 100;
	m_Para.MinSpeedHold = 10;
	
	//pCtlPtList = new CtlPointList();
}

/*----------------------------------------------------------------------------*/

BahnDesign::~BahnDesign(void)
{
//	delete pCtlPtList;
}

/*----------------------------------------------------------------------------*/
tagBahnDesignPt* BahnDesign::add()
{
	tagBahnDesignPt *pPt=&(Pts[Count++]);
	pPt->time=0;
	pPt->Speed=0;
	pPt->Vect.alpha=0;
    return pPt;
}

/*----------------------------------------------------------------------------*/
tagBahnDesignPt * BahnDesign::insert(int id)
{
    tagBahnDesignPt * pPt=NULL;

    if ((id>=0)&&(id<=Count)&&(Count<MaxBahnDesignPtCount)) 
	{
        for (int i=id;i<Count;i++) 
		{
            Pts[i+1]=Pts[i];
        }

        pPt = &(Pts[id]);
        Count++;
    }

    return pPt;
}

/*----------------------------------------------------------------------------*/
void BahnDesign::remove(int id)
{
    if ((id>=0)&&(id<=Count)&&(Count<MaxBahnDesignPtCount)) 
	{
        for (int i=id;i<Count-1;i++) 
		{
            Pts[i] = Pts[i+1];
        }

        Count--;
    }
}

/*----------------------------------------------------------------------------*/
/*
 * Search a BahnPoint or its arrow
 */
int BahnDesign::SearchBDPt(double x, double y, double r, double tArrowLen, bool IsSearchArrow)
{
	int i;

	for (i=0;i<Count;i++) 
	{
		tagVect vect = Pts[i].Vect;
		double x1 = vect.x;
		double y1 = vect.y;

		if (IsSearchArrow) 
		{
			x1 += tArrowLen*cos(vect.alpha);
			y1 += tArrowLen*sin(vect.alpha);
		}

		double dx = x-x1;
		double dy = y-y1;

		if ((fabs(dx)<r)&&(fabs(dy)<r))
		{
			break;
		}
	}
	
	if (i<Count)
	{
		return i;
	}
	else
	{
		return -1;
	}
}

/*----------------------------------------------------------------------------*/
void BahnDesign::CalcGeoI_far(tagCalcGeoIContext &gc)
{
	tagVect &v1 = gc.p1->Vect;
	tagVect &v2 = gc.p2->Vect;
	tagBahnStrecke &stk = gc.stk;

	tagCircle &c1 = gc.c[gc.ci1];
	tagCircle &c2 = gc.c[gc.ci2];

	bool isLeft1 = (gc.ci1%2==0);
	bool isLeft2 = (gc.ci2%2==0);

	double alpha = atan2(c2.oy-c1.oy, c2.ox-c1.ox);
	double alpha_nv1, alpha_nv2;

	gc.p1->StkCount = 0;
	gc.p1->StkIdSt = StkCt;

	if (isLeft1 != isLeft2) 
	{ // cross
		double l = gc.dist_c1_c2 / 2;
		double sg = isLeft1 ? -1 : 1;
		double delta = _fmod(asin(c1.r / l), _2pi) * sg;
		alpha += delta;
		alpha_nv1 = alpha - sg * pi / 2;
		alpha_nv2 = alpha + sg * pi / 2;
	}
	else 
	{ // parallel
		double sg = isLeft2 ? -1 : 1;
		alpha = atan2(c2.oy - c1.oy, c2.ox - c1.ox);
		alpha_nv1 = alpha - sg * pi / 2;
		alpha_nv2 = alpha_nv1;
	}

	gc.tv1.x = c1.ox + c1.r * cos(alpha_nv1);
	gc.tv1.y = c1.oy + c1.r * sin(alpha_nv1);
	gc.tv2.x = c2.ox + c2.r * cos(alpha_nv2);
	gc.tv2.y = c2.oy + c2.r * sin(alpha_nv2);
	gc.tv1.alpha = alpha;
	gc.tv2.alpha = alpha;
	gc.arc1.c = c1;
	gc.arc1.a = _fmod(atan2(v1.y - c1.oy, v1.x - c1.ox), _2pi);
	gc.arc1.alen = gc.arc1.a - alpha_nv1;
	gc.arc2.c = c2;
	gc.arc2.a = alpha_nv2;
	gc.arc2.alen = gc.arc2.a - atan2(v2.y - c2.oy, v2.x - c2.ox);

	if ((v1.x!=gc.tv1.x)&&(v1.y!=gc.tv1.y)) 
	{
		stk.Type = 1; // circle
		stk.VectSt = gc.p1->Vect;
		stk.arc = gc.arc1;
		
		if (isLeft1&&(stk.arc.alen<0))
		{
			stk.arc.alen += _2pi;
		}
		
		if (!isLeft1&&(stk.arc.alen>0))
		{
			stk.arc.alen -= _2pi;
		}
		
		gc.arc1.alen = stk.arc.alen;
		stk.length = fabs(stk.arc.alen * stk.arc.c.r);
//		stk.EndPt.time = -1;
		stk.VectEd = gc.tv1;
		Stks[StkCt++] = stk;
		stk.VectSt = stk.VectEd;
		
		gc.p1->StkCount++;
	}

	stk.Type = 0; // line
	//stk.VectEd = stk.VectSt;
	//stk.EndPt.time = -1;
	stk.VectEd = gc.tv2;
	stk.length = getDist(gc.tv1.x,gc.tv1.y,gc.tv2.x,gc.tv2.y);
	Stks[StkCt++] = stk;
	gc.p1->StkCount++;
	stk.VectSt = stk.VectEd;

	if ((v2.x!=gc.tv2.x)&&(v2.y!=gc.tv2.y)) 
	{
		stk.Type = 1; // circle
		stk.arc = gc.arc2;
		if (isLeft2&&(stk.arc.alen<0))
			stk.arc.alen += _2pi;
		if (!isLeft2&&(stk.arc.alen>0))
			stk.arc.alen -= _2pi;
		gc.arc2.alen = stk.arc.alen;
		stk.length = fabs(stk.arc.alen * stk.arc.c.r);
		stk.VectEd = v2;
		Stks[StkCt++] = stk;
		gc.p1->StkCount++;
	}

}

/*----------------------  use far mode when possible  ------------------------*/
// return true when CalcGeoI_far is called
bool BahnDesign::CheckDistance(tagCalcGeoIContext &gc)
{
	double d[4];
	// 0    1    2    3
	// cl1, cr1, cl2, cr2;
	tagVect &v1=gc.p1->Vect, &v2=gc.p2->Vect;
	
	getVectTanCircle(v1, gc.c[0], gc.c[1], m_Para.MaxR);  // left & right tangent circle of v1
	getVectTanCircle(v2, gc.c[2], gc.c[3], m_Para.MaxR);  // left & right tangent circle of v2
	
	d[0]=getDist(gc.c[0].ox, gc.c[0].oy,gc.c[2].ox,gc.c[2].oy);
	d[1]=getDist(gc.c[0].ox, gc.c[0].oy,gc.c[3].ox,gc.c[3].oy);
	d[2]=getDist(gc.c[1].ox, gc.c[1].oy,gc.c[2].ox,gc.c[2].oy);
	d[3]=getDist(gc.c[1].ox, gc.c[1].oy,gc.c[3].ox,gc.c[3].oy);
	
	int min_id=-1;

	for (int i=0;i<4;i++) {
		if ((d[i]>=2*m_Para.MaxR)||(i==0)||(i==3)) {
			if (min_id<0) {
				min_id=i;
			}
			else {
				if (d[i]<d[min_id]) min_id=i;
			}
		}
	}

	if (min_id<0) return false;
	
/*
	for (int i=1;i<4;i++) {
		if (((i==1)||(i==2))&&(d[i]<2*RMax)) continue;
		if (d[i]<d[min_id]) min_id=i;
	}
*/
	//if (d[min_id]>=2*RMax) {
		gc.dist_c1_c2 = d[min_id];

		switch (min_id) 
		{
		case 0:
			gc.ci1=0;
			gc.ci2=2;
			break;
		case 1: 
			gc.ci1=0;
			gc.ci2=3;
			break;
		case 2: 
			gc.ci1=1;
			gc.ci2=2;
			break;
		case 3: 
			gc.ci1=1;
			gc.ci2=3;
			break;
		}

		//int OldStkCt=this->StkCt;
		CalcGeoI_far(gc);
		/*
		bool retry=false;
		if (fabs(gc.arc1.alen)>1.5*pi) {
			gc.ci1=(gc.ci1+1)%2;
			retry=true;
		}
		if (fabs(gc.arc2.alen)>1.5*pi) {
			gc.ci2=(gc.ci2+1)%2+2;
			retry=true;
		}
		if (retry) {
			this->StkCt=OldStkCt;
			CalcGeoI_far(gc);
		}

		*/
		return true;
	//}
	//return false;
}


/*----------------------------------------------------------------------------*/
void BahnDesign::CalcGeoI(tagCalcGeoIContext &gc)
{
	double d, l;
	tagVect &v1 = gc.p1->Vect;
	tagVect	&v2 = gc.p2->Vect;
	tagVect &vc = gc.vc;
	tagBahnStrecke &stk = gc.stk;

    v1.alpha = _fmod(v1.alpha, _2pi);
    v2.alpha = _fmod(v2.alpha, _2pi);

	//  use far mode when possible
	if (CheckDistance(gc))
	{
		return;
	}

	gc.isParallel = !getCrossVect(v1, v2, vc);
	if (gc.isParallel) // pararell
	{
		return;
	}
	
	/*
	stk.arc.alen = _fmod(v1.alpha-v2.alpha,_2pi);
	if (stk.arc.alen>pi)
		stk.arc.alen-=_2pi;
	*/
	
    double &d1 = gc.d1, &d2 = gc.d2, &dmin = gc.dmin;
	d1 = sqrt(pow(v1.x - vc.x, 2) + pow(v1.y - vc.y, 2));
    d2 = sqrt(pow(v2.x - vc.x, 2) + pow(v2.y - vc.y, 2));
    dmin= fabs(m_Para.MinR / tan(vc.alpha - v2.alpha));
	
	// calc CrossMode
	//enCrossMode CrossMode = CM_Easy;
	int &CrossMode = gc.CrossMode;
	CrossMode = CM_Easy;
    if (((vc.x-v1.x)*cos(v1.alpha)<0)||((vc.y-v1.y)*sin(v1.alpha)<0))
	{
        CrossMode+=1;
    }
    else
	{
        if (d1<dmin)
		{
			CrossMode+=1;
		}
    }

    if (((v2.x - vc.x) * cos(v2.alpha) < 0) || ((v2.y - vc.y) * sin(v2.alpha) < 0)){
        CrossMode += 2;
    }
    else{
        if (d2<dmin)
		{
			CrossMode += 2;
		}
    }

	// easy mode
	if (CrossMode==CM_Easy) 
	{
		d = (d1<d2) ? d1 : d2;
		stk.arc.c.r = d / tan(fabs(stk.arc.alen) / 2);
		l = sqrt(stk.arc.c.r * stk.arc.c.r + d * d);
		stk.arc.c.ox = vc.x + l * cos(vc.alpha);
		stk.arc.c.oy = vc.y + l * sin(vc.alpha);
		stk.VectSt = gc.p1->Vect;

		if (d1>d2) 
		{
			// line
			stk.Type = 0; // line
			stk.length = d1 - d2;
			Stks[StkCt++] = stk;
			stk.VectSt.x += stk.length * cos(v1.alpha);
			stk.VectSt.y += stk.length * sin(v1.alpha);
		}

		// kreis
		stk.Type = 1; // circle
		stk.length = stk.arc.c.r * stk.arc.alen;

		if (stk.arc.alen>0) 
		{
			stk.arc.a=_fmod(v1.alpha+pi/2,_2pi);
		}
		else
		{
			stk.arc.a=_fmod(v1.alpha-pi/2,_2pi);
		}

		Stks[StkCt++] = stk;

		if (d1<d2) 
		{
			// line
			stk.Type = 0; // line
			stk.length = d2 - d1;
			stk.VectSt = gc.p2->Vect;
			stk.VectSt.x = v2.x - stk.length * cos(v2.alpha);
			stk.VectSt.y = v2.y - stk.length * sin(v2.alpha);
			Stks[StkCt++] = stk;
		}
	}
}

/*----------------------  Geometry part of the Calculation  ------------------*/
void BahnDesign::CalcGeo(void)
{
    if (Count>1) 
    {
        for (int i=0;i<Count-1;i++) 
        {
          gc.p1=&(Pts[i]);
          gc.p2=&(Pts[i+1]);
          CalcGeoI(gc);
        }
    }
}

/*----------------------------------------------------------------------------*/
/*
 * Description: method to calculate the speed 
 * Para: none
 * Return: none
 */
void BahnDesign::CalcSpeed(void)
{
	if (Count<2)
	{
		return;
	}
	
	SpdList.Reset();

	for (int i=0; i<Count-1; i++)
	{
		SpdList.Append(&(Pts[i]), &(Stks[Pts[i].StkIdSt]), i);
	}
}

/*----------------------------------------------------------------------------*/
/*
 * Description: method to calculate
 * Para: none
 * Return: none
 */
void BahnDesign::Calc(void)
{
	StkCt = 0; // Count of Strecke
	CalcGeo(); // geometric
	CalcSpeed(); // Speed

	CtlPtList.Reset(Stks, StkCt, SpdList);

	// debug
	int i;

	if (m_Log.is_open())
	{
		m_Log.close();
	}

	m_Log.open("dump.log");

	m_Log << endl << endl << "=========== Dump DesignPts ========" << endl;
	
	for (i=0;i<Count;i++)
	{
		_dump(Pts[i], i);
		m_Log << endl;
	}


	m_Log << endl << endl << "============ Dump Strecke =========" << endl;
	
	for (i=0;i<StkCt;i++)
	{
		_dump(Stks[i], i);
		m_Log << endl;
	}


	m_Log << endl << endl << "============ Dump speeds ==========" << endl;

	for (i=0;i<SpdList.GetCount();i++)
	{
		_dump(SpdList.Item(i), i);
		m_Log << endl;
	}


	m_Log << endl << endl << "=========== Dump controls =========" << endl;

	for (i=0;i<CtlPtList.GetCount();i++)
	{
		_dump(CtlPtList.Item(i), i);
		m_Log << endl;
	}
	
	m_Log.close();
}
